import typing

from ..data_types import Fingerprint
from ..data_types import Transaction


class ExtractorBase:
    def __init__(self, input_file: typing.TextIO):
        self.input_file = input_file

    def detect(self) -> bool:
        raise NotImplementedError()

    def fingerprint(self) -> Fingerprint | None:
        raise NotImplementedError()

    def __call__(self) -> typing.Generator[Transaction, None, None]:
        raise NotImplementedError()



import datetime
import decimal
import functools
import pathlib

import pytest

from beanhub_extract.data_types import Fingerprint
from beanhub_extract.data_types import Transaction
from beanhub_extract.extractors.wealthsimple import WealthsimpleExtractor
from beanhub_extract.extractors.wealthsimple import parse_date
from beanhub_extract.utils import strip_txn_base_path


@pytest.mark.parametrize(
    "date_str, expected",
    [
        ("2024-04-01", datetime.date(2024, 4, 1)),
    ],
)
def test_parse_date(date_str: str, expected: datetime.date):
    assert parse_date(date_str) == expected


@pytest.mark.parametrize(
    "input_file, expected",
    [
        (
            "wealthsimple.csv",
            [
                Transaction(
                    extractor="wealthsimple",
                    file="wealthsimple.csv",
                    lineno=1,
                    reversed_lineno=-5,
                    date=datetime.date(2024, 4, 1),
                    desc="Interest earned",
                    amount=decimal.Decimal("12.34"),
                    type="INT",
                    extra={"balance": decimal.Decimal("5123.45")},
                ),
                Transaction(
                    extractor="wealthsimple",
                    file="wealthsimple.csv",
                    lineno=2,
                    reversed_lineno=-4,
                    date=datetime.date(2024, 4, 2),
                    desc="Coffee Shop",
                    amount=decimal.Decimal("-4.56"),
                    type="SPEND",
                    extra={"balance": decimal.Decimal("5118.89")},
                ),
                Transaction(
                    extractor="wealthsimple",
                    file="wealthsimple.csv",
                    lineno=3,
                    reversed_lineno=-3,
                    date=datetime.date(2024, 4, 3),
                    desc="Cash back reward",
                    amount=decimal.Decimal("1.23"),
                    type="CASHBACK",
                    extra={"balance": decimal.Decimal("5120.12")},
                ),
                Transaction(
                    extractor="wealthsimple",
                    file="wealthsimple.csv",
                    lineno=4,
                    reversed_lineno=-2,
                    date=datetime.date(2024, 4, 4),
                    desc="Grocery Store",
                    amount=decimal.Decimal("-45.67"),
                    type="SPEND",
                    extra={"balance": decimal.Decimal("5074.45")},
                ),
                Transaction(
                    extractor="wealthsimple",
                    file="wealthsimple.csv",
                    lineno=5,
                    reversed_lineno=-1,
                    date=datetime.date(2024, 4, 15),
                    desc="Direct deposit",
                    amount=decimal.Decimal("1234.56"),
                    type="AFT_IN",
                    extra={"balance": decimal.Decimal("6309.01")},
                ),
            ],
        ),
    ],
)

def test_wealthsimple_extractor(
    fixtures_folder: pathlib.Path, input_file: str, expected: list[Transaction]
):
    with open(fixtures_folder / input_file, "rt", encoding="utf-8") as fo:
        extractor = WealthsimpleExtractor(fo)
        assert (
            list(
                map(
                    functools.partial(strip_txn_base_path, fixtures_folder), extractor()
                )
            )
            == expected
        )


@pytest.mark.parametrize(
    "input_file, expected",
    [
        ("wealthsimple.csv", True),
        ("mercury.csv", False),
        ("empty.csv", False),
        ("other.csv", False),
        (pytest.lazy_fixture("zip_file"), False),
    ],
)
def test_wealthsimple_detect(
    fixtures_folder: pathlib.Path, input_file: str, expected: bool
):
    with open(fixtures_folder / input_file, "rt", encoding="utf-8") as fo:
        extractor = WealthsimpleExtractor(fo)
        assert extractor.detect() == expected


def test_wealthsimple_fingerprint(fixtures_folder: pathlib.Path):
    with open(fixtures_folder / "wealthsimple.csv", "rt", encoding="utf-8") as fo:
        extractor = WealthsimpleExtractor(fo)
        assert extractor.fingerprint() == Fingerprint(
            starting_date=datetime.date(2024, 4, 1),
            first_row_hash="f9d844138962e483beecd7f44ada98f0995810e4b1e3fc6d3b27d1b6189318b6",
        )



import datetime
import decimal
import functools
import pathlib

import pytest

from beanhub_extract.data_types import Fingerprint
from beanhub_extract.data_types import Transaction
from beanhub_extract.extractors.csv import CSVExtractor
from beanhub_extract.utils import strip_txn_base_path


@pytest.mark.parametrize(
    "input_file, expected",
    [
        (
            "csv.csv",
            [
                Transaction(
                    extractor="csv",
                    file="csv.csv",
                    lineno=1,
                    reversed_lineno=-2,
                    transaction_id="4c5484e6-dad0-4243-8c52-be74d6c84d59",
                    date=datetime.date(2025, 6, 28),
                    post_date=datetime.date(2025, 7, 1),
                    timestamp=datetime.datetime.fromisoformat(
                        "2025-06-29T00:32:53.036910+00:00"
                    ),
                    timezone="America/Los_Angeles",
                    desc="BeanHub subscription",
                    bank_desc="BEANHUB SUB",
                    amount=decimal.Decimal("12.34"),
                    currency="USD",
                    category="Business",
                    subcategory="Software",
                    pending=False,
                    status="Paid",
                    type="Debit",
                    source_account="Mercury",
                    payee="BeanHub",
                    gl_code="5100",
                    name_on_card="Fang-Pen Lin",
                    last_four_digits="1234",
                    extra=dict(_custom="val0"),
                ),
                Transaction(
                    extractor="csv",
                    file="csv.csv",
                    lineno=2,
                    reversed_lineno=-1,
                    transaction_id="39751802-7646-4c24-b9bb-227c9f68a8f2",
                    date=datetime.date(2025, 7, 3),
                    timestamp=datetime.datetime.fromisoformat(
                        "2025-07-03T00:48:51.024820+00:00"
                    ),
                    timezone="America/Los_Angeles",
                    desc="CakeLens subscription",
                    bank_desc="CAKELENS",
                    amount=decimal.Decimal("5.67"),
                    currency="USD",
                    category="Business",
                    subcategory="Software",
                    pending=True,
                    status="Paid",
                    type="Debit",
                    source_account="Mercury",
                    note="AI gen video detection service",
                    payee="BeanHub",
                    gl_code="5100",
                    name_on_card="Fang-Pen Lin",
                    last_four_digits="1234",
                    extra=dict(_custom="val1"),
                ),
            ],
        ),
    ],
)
def test_extractor(
    fixtures_folder: pathlib.Path, input_file: str, expected: list[Transaction]
):
    with open(fixtures_folder / input_file, "rt") as fo:
        extractor = CSVExtractor(fo)
        assert (
            list(
                map(
                    functools.partial(strip_txn_base_path, fixtures_folder), extractor()
                )
            )
            == expected
        )


def test_fingerprint(fixtures_folder: pathlib.Path):
    with open(fixtures_folder / "csv.csv", "rt") as fo:
        extractor = CSVExtractor(fo)
        assert extractor.fingerprint() == Fingerprint(
            starting_date=datetime.date(2025, 6, 28),
            first_row_hash="3128963b757e527f3d192edbecf4ee55f22abbaa3736b402bec93b25a5aae458",
        )



import datetime
import decimal
import functools
import pathlib

import pytest
import pytz

from beanhub_extract.data_types import Fingerprint
from beanhub_extract.data_types import Transaction
from beanhub_extract.extractors.mercury import MercuryExtractor
from beanhub_extract.extractors.mercury import parse_date
from beanhub_extract.extractors.mercury import parse_datetime
from beanhub_extract.extractors.mercury import parse_time
from beanhub_extract.utils import strip_txn_base_path


@pytest.mark.parametrize(
    "date_str, expected",
    [
        ("05-04-2024", datetime.date(2024, 5, 4)),
    ],
)
def test_parse_date(date_str: str, expected: datetime.date):
    assert parse_date(date_str) == expected


@pytest.mark.parametrize(
    "time_str, expected",
    [
        ("14:35:37", datetime.time(14, 35, 37)),
    ],
)
def test_parse_time(time_str: str, expected: datetime.time):
    assert parse_time(time_str) == expected


@pytest.mark.parametrize(
    "datetime_str, expected",
    [
        ("05-04-2024 14:35:37", datetime.datetime(2024, 5, 4, 14, 35, 37)),
    ],
)
def test_parse_datetime(datetime_str: str, expected: datetime.datetime):
    assert parse_datetime(datetime_str) == expected


@pytest.mark.parametrize(
    "input_file, expected",
    [
        (
            "mercury.csv",
            [
                Transaction(
                    extractor="mercury",
                    file="mercury.csv",
                    lineno=1,
                    reversed_lineno=-4,
                    date=datetime.date(2024, 4, 17),
                    timestamp=datetime.datetime(
                        2024, 4, 17, 21, 30, 40, tzinfo=pytz.UTC
                    ),
                    timezone="UTC",
                    desc="GUSTO",
                    bank_desc="GUSTO; FEE 111111; Launch Platform LLC",
                    amount=decimal.Decimal("-46.00"),
                    category="",
                    status="Sent",
                    source_account="Mercury Checking xx12",
                    note="",
                    reference="",
                    currency="",
                    gl_code="",
                    name_on_card="",
                    last_four_digits="",
                ),
                Transaction(
                    extractor="mercury",
                    file="mercury.csv",
                    lineno=2,
                    reversed_lineno=-3,
                    date=datetime.date(2024, 4, 16),
                    timestamp=datetime.datetime(
                        2024, 4, 16, 3, 25, 55, tzinfo=pytz.UTC
                    ),
                    timezone="UTC",
                    desc="Amazon Web Services",
                    bank_desc="Amazon web services",
                    amount=decimal.Decimal("-353.63"),
                    category="Software",
                    status="Sent",
                    source_account="Mercury Credit",
                    note="",
                    reference="",
                    currency="USD",
                    gl_code="",
                    name_on_card="Fang-Pen Lin",
                    last_four_digits="5678",
                ),
                Transaction(
                    extractor="mercury",
                    file="mercury.csv",
                    lineno=3,
                    reversed_lineno=-2,
                    date=datetime.date(2024, 4, 16),
                    timestamp=datetime.datetime(
                        2024, 4, 16, 3, 24, 57, tzinfo=pytz.UTC
                    ),
                    timezone="UTC",
                    desc="Adobe",
                    bank_desc="ADOBE  *ADOBE",
                    amount=decimal.Decimal("-54.99"),
                    category="Software",
                    status="Sent",
                    source_account="Mercury Credit",
                    note="",
                    reference="",
                    currency="USD",
                    gl_code="",
                    name_on_card="Fang-Pen Lin",
                    last_four_digits="5678",
                ),
                Transaction(
                    extractor="mercury",
                    file="mercury.csv",
                    lineno=4,
                    reversed_lineno=-1,
                    date=datetime.date(2024, 4, 15),
                    timestamp=datetime.datetime(
                        2024, 4, 15, 14, 35, 37, tzinfo=pytz.UTC
                    ),
                    timezone="UTC",
                    desc="Jane Doe",
                    bank_desc="Send Money transaction initiated on Mercury",
                    amount=decimal.Decimal("-1500.00"),
                    category="",
                    status="Sent",
                    source_account="Mercury Checking xx1234",
                    note="",
                    reference="Profit distribution",
                    currency="",
                    gl_code="",
                    name_on_card="",
                    last_four_digits="",
                ),
            ],
        ),
    ],
)
def test_extractor(
    fixtures_folder: pathlib.Path, input_file: str, expected: list[Transaction]
):
    with open(fixtures_folder / input_file, "rt") as fo:
        extractor = MercuryExtractor(fo)
        assert (
            list(
                map(
                    functools.partial(strip_txn_base_path, fixtures_folder), extractor()
                )
            )
            == expected
        )


@pytest.mark.parametrize(
    "input_file, expected",
    [
        ("mercury.csv", True),
        ("chase_credit_card.csv", False),
        ("empty.csv", False),
        ("other.csv", False),
        (pytest.lazy_fixture("zip_file"), False),
    ],
)
def test_detect(fixtures_folder: pathlib.Path, input_file: str, expected: bool):
    with open(fixtures_folder / input_file, "rt") as fo:
        extractor = MercuryExtractor(fo)
        assert extractor.detect() == expected


def test_fingerprint(fixtures_folder: pathlib.Path):
    with open(fixtures_folder / "mercury.csv", "rt") as fo:
        extractor = MercuryExtractor(fo)
        assert extractor.fingerprint() == Fingerprint(
            starting_date=datetime.date(2024, 4, 15),
            first_row_hash="24039f29eb959f67b19e1bdd0b8466513bef2c5bcdf5c11dd463e744abc23f85",
        )



import datetime
import decimal
import functools
import pathlib

import pytest

from beanhub_extract.data_types import Fingerprint
from beanhub_extract.data_types import Transaction
from beanhub_extract.extractors.chase import ChaseCreditCardExtractor
from beanhub_extract.extractors.chase import parse_date
from beanhub_extract.utils import strip_txn_base_path


@pytest.mark.parametrize(
    "date_str, expected",
    [
        ("05/04/2024", datetime.date(2024, 5, 4)),
    ],
)
def test_parse_date(date_str: str, expected: datetime.date):
    assert parse_date(date_str) == expected


@pytest.mark.parametrize(
    "input_file, expected",
    [
        (
            "chase_credit_card.csv",
            [
                Transaction(
                    extractor="chase_credit_card",
                    file="chase_credit_card.csv",
                    lineno=1,
                    reversed_lineno=-5,
                    date=datetime.date(2024, 4, 9),
                    post_date=datetime.date(2024, 4, 9),
                    desc="AUTOMATIC PAYMENT - THANK",
                    amount=decimal.Decimal("123.45"),
                    category="",
                    type="Payment",
                    note="",
                ),
                Transaction(
                    extractor="chase_credit_card",
                    file="chase_credit_card.csv",
                    lineno=2,
                    reversed_lineno=-4,
                    date=datetime.date(2024, 4, 3),
                    post_date=datetime.date(2024, 4, 5),
                    desc="APPLE.COM/BILL",
                    amount=decimal.Decimal("-1.23"),
                    category="Shopping",
                    type="Sale",
                    note="",
                ),
                Transaction(
                    extractor="chase_credit_card",
                    file="chase_credit_card.csv",
                    lineno=3,
                    reversed_lineno=-3,
                    date=datetime.date(2024, 4, 2),
                    post_date=datetime.date(2024, 4, 3),
                    desc="COSTCO WHSE #01234",
                    amount=decimal.Decimal("-4.56"),
                    category="Shopping",
                    type="Sale",
                    note="",
                ),
                Transaction(
                    extractor="chase_credit_card",
                    file="chase_credit_card.csv",
                    lineno=4,
                    reversed_lineno=-2,
                    date=datetime.date(2024, 4, 2),
                    post_date=datetime.date(2024, 4, 3),
                    desc="Amazon web services",
                    amount=decimal.Decimal("-6.54"),
                    category="Personal",
                    type="Sale",
                    note="",
                ),
                Transaction(
                    extractor="chase_credit_card",
                    file="chase_credit_card.csv",
                    lineno=5,
                    reversed_lineno=-1,
                    date=datetime.date(2024, 4, 1),
                    post_date=datetime.date(2024, 4, 2),
                    desc="GITHUB  INC.",
                    amount=decimal.Decimal("-4.00"),
                    category="Professional Services",
                    type="Sale",
                    note="",
                ),
            ],
        ),
    ],
)
def test_credit_card_extractor(
    fixtures_folder: pathlib.Path, input_file: str, expected: list[Transaction]
):
    with open(fixtures_folder / input_file, "rt") as fo:
        extractor = ChaseCreditCardExtractor(fo)
        assert (
            list(
                map(
                    functools.partial(strip_txn_base_path, fixtures_folder), extractor()
                )
            )
            == expected
        )


@pytest.mark.parametrize(
    "input_file, expected",
    [
        ("chase_credit_card.csv", True),
        ("mercury.csv", False),
        ("empty.csv", False),
        ("other.csv", False),
        (pytest.lazy_fixture("zip_file"), False),
    ],
)
def test_credit_card_detect(
    fixtures_folder: pathlib.Path, input_file: str, expected: bool
):
    with open(fixtures_folder / input_file, "rt") as fo:
        extractor = ChaseCreditCardExtractor(fo)
        assert extractor.detect() == expected


def test_credit_card_fingerprint(fixtures_folder: pathlib.Path):
    with open(fixtures_folder / "chase_credit_card.csv", "rt") as fo:
        extractor = ChaseCreditCardExtractor(fo)
        assert extractor.fingerprint() == Fingerprint(
            starting_date=datetime.date(2024, 4, 1),
            first_row_hash="f68f953ef2d7c7a088924728b8b6b573120fcb384ff3fbd4ca382335a336acc3",
        )



import pathlib

import pytest

from beanhub_extract.data_types import Transaction
from beanhub_extract.utils import strip_base_path
from beanhub_extract.utils import strip_txn_base_path


@pytest.mark.parametrize(
    "base_path, input_path, expected",
    [
        ("/path/to/", "/path/to/nested/my-file.csv", "nested/my-file.csv"),
    ],
)
def test_strip_base_path(base_path: str, input_path: str, expected: str):
    assert (
        strip_base_path(
            pathlib.PurePosixPath(base_path), pathlib.PurePosixPath(input_path)
        )
        == expected
    )


@pytest.mark.parametrize(
    "base_path, txn, expected",
    [
        (
            "/path/to/",
            Transaction(extractor="", file="/path/to/nested/my-file.csv"),
            Transaction(extractor="", file="nested/my-file.csv"),
        ),
    ],
)
def test_strip_txn_base_path(base_path: str, txn: Transaction, expected: Transaction):
    assert (
        strip_txn_base_path(pathlib.PurePosixPath(base_path), txn, pure_posix=True)
        == expected
    )



import datetime

import pytest

from beanhub_extract.utils import parse_date


@pytest.mark.parametrize(
    "date_str, expected",
    [
        ("2024-05-04", datetime.date(2024, 5, 4)),
    ],
)
def test_parse_date(date_str: str, expected: datetime.date):
    assert parse_date(date_str) == expected



import csv
import datetime
import decimal
import hashlib
import os
import typing

from ..data_types import Fingerprint
from ..data_types import Transaction
from .base import ExtractorBase


def parse_date(date_str: str) -> datetime.date:
    parts = date_str.split("/")
    return datetime.date(int(parts[-1]), *(map(int, parts[:-1])))


class ChaseCreditCardExtractor(ExtractorBase):
    EXTRACTOR_NAME = "chase_credit_card"
    DEFAULT_IMPORT_ID = "{{ file | as_posix_path }}:{{ reversed_lineno }}"
    ALL_FIELDS = [
        "Transaction Date",
        "Post Date",
        "Description",
        "Category",
        "Type",
        "Amount",
        "Memo",
    ]

    def detect(self) -> bool:
        reader = csv.DictReader(self.input_file)
        try:
            return reader.fieldnames == self.ALL_FIELDS
        except Exception:
            return False

    def fingerprint(self) -> Fingerprint | None:
        reader = csv.DictReader(self.input_file)
        row = None
        for row in reader:
            pass
        if row is None:
            return
        hash = hashlib.sha256()
        for field in reader.fieldnames:
            hash.update(row[field].encode("utf8"))
        return Fingerprint(
            starting_date=parse_date(row["Transaction Date"]),
            first_row_hash=hash.hexdigest(),
        )

    def __call__(self) -> typing.Generator[Transaction, None, None]:
        filename = None
        if hasattr(self.input_file, "name"):
            filename = self.input_file.name
        row_count_reader = csv.DictReader(self.input_file)
        row_count = 0
        for _ in row_count_reader:
            row_count += 1
        self.input_file.seek(os.SEEK_SET, 0)
        reader = csv.DictReader(self.input_file)
        for i, row in enumerate(reader):
            kwargs = dict(
                date=parse_date(row.pop("Transaction Date")),
                post_date=parse_date(row.pop("Post Date")),
                desc=row.pop("Description"),
                category=row.pop("Category"),
                type=row.pop("Type"),
                amount=decimal.Decimal(row.pop("Amount")),
                note=row.pop("Memo"),
            )
            if row:
                kwargs["extra"] = row

            yield Transaction(
                extractor=self.EXTRACTOR_NAME,
                file=filename,
                lineno=i + 1,
                reversed_lineno=i - row_count,
                **kwargs
            )



import csv
import datetime
import decimal
import hashlib
import os
import typing

import pytz

from ..data_types import Fingerprint
from ..data_types import Transaction
from .base import ExtractorBase


def parse_date(date_str: str) -> datetime.date:
    parts = date_str.split("-")
    return datetime.date(int(parts[-1]), *(map(int, parts[:-1])))


def parse_time(time_str: str) -> datetime.time:
    parts = time_str.split(":")
    return datetime.time(*(map(int, parts)))


def parse_datetime(timestamp_str: str) -> datetime.datetime:
    parts = timestamp_str.split(" ")
    date = parse_date(parts[0])
    time = parse_time(parts[1])
    return datetime.datetime.combine(date, time)


class MercuryExtractor(ExtractorBase):
    EXTRACTOR_NAME = "mercury"
    DEFAULT_IMPORT_ID = "{{ file | as_posix_path }}:{{ reversed_lineno }}"
    ALL_FIELDS = [
        "Date (UTC)",
        "Description",
        "Amount",
        "Status",
        "Source Account",
        "Bank Description",
        "Reference",
        "Note",
        "Last Four Digits",
        "Name On Card",
        "Category",
        "GL Code",
        "Timestamp",
        "Original Currency",
    ]

    def detect(self) -> bool:
        reader = csv.DictReader(self.input_file)
        try:
            return reader.fieldnames == self.ALL_FIELDS
        except Exception:
            return False

    def fingerprint(self) -> Fingerprint | None:
        reader = csv.DictReader(self.input_file)
        row = None
        for row in reader:
            pass
        if row is None:
            return
        hash = hashlib.sha256()
        for field in reader.fieldnames:
            hash.update(row[field].encode("utf8"))
        return Fingerprint(
            starting_date=parse_date(row["Date (UTC)"]),
            first_row_hash=hash.hexdigest(),
        )

    def __call__(self) -> typing.Generator[Transaction, None, None]:
        filename = None
        if hasattr(self.input_file, "name"):
            filename = self.input_file.name
        row_count_reader = csv.DictReader(self.input_file)
        row_count = 0
        for _ in row_count_reader:
            row_count += 1
        self.input_file.seek(os.SEEK_SET, 0)
        reader = csv.DictReader(self.input_file)
        timezone = pytz.UTC
        for i, row in enumerate(reader):
            kwargs = dict(
                date=parse_date(row.pop("Date (UTC)")),
                desc=row.pop("Description"),
                amount=decimal.Decimal(row.pop("Amount")),
                status=row.pop("Status"),
                source_account=row.pop("Source Account"),
                bank_desc=row.pop("Bank Description"),
                reference=row.pop("Reference"),
                note=row.pop("Note"),
                category=row.pop("Category"),
                currency=row.pop("Original Currency"),
                name_on_card=row.pop("Name On Card"),
                last_four_digits=row.pop("Last Four Digits"),
                gl_code=row.pop("GL Code"),
                timestamp=timezone.localize(parse_datetime(row.pop("Timestamp"))),
            )
            if row:
                kwargs["extra"] = row

            yield Transaction(
                extractor=self.EXTRACTOR_NAME,
                file=filename,
                lineno=i + 1,
                reversed_lineno=i - row_count,
                timezone="UTC",
                **kwargs
            )



import csv
import datetime
import decimal
import hashlib
import typing

from ..data_types import Fingerprint
from ..data_types import Transaction
from .base import ExtractorBase


def parse_date(date_str: str) -> datetime.date:
    parts = date_str.split("-")
    return datetime.date(int(parts[0]), int(parts[1]), int(parts[2]))


class WealthsimpleExtractor(ExtractorBase):
    EXTRACTOR_NAME = "wealthsimple"
    DEFAULT_IMPORT_ID = "{{ file | as_posix_path }}:{{ reversed_lineno }}"
    ALL_FIELDS = [
        "date",
        "transaction",
        "description",
        "amount",
        "balance",
    ]

    def detect(self) -> bool:
        reader = csv.DictReader(self.input_file)
        try:
            return reader.fieldnames == self.ALL_FIELDS
        except Exception:
            return False

    def fingerprint(self) -> Fingerprint | None:
        self.input_file.seek(0)
        reader = csv.DictReader(self.input_file)
        try:
            row = next(reader)
        except StopIteration:
            return None
        
        hash = hashlib.sha256()
        for field in reader.fieldnames:
            hash.update(row[field].encode("utf8"))
        
        return Fingerprint(
            starting_date=parse_date(row["date"]),
            first_row_hash=hash.hexdigest(),
        )

    def __call__(self) -> typing.Generator[Transaction, None, None]:
        self.input_file.seek(0)
        reader = csv.DictReader(self.input_file)
        rows = list(reader)
        row_count = len(rows)
        
        filename = getattr(self.input_file, "name", None)
        
        for i, row in enumerate(rows):
            date_str = row["date"]
            date = parse_date(date_str)
            transaction_type = row["transaction"]
            amount = decimal.Decimal(row["amount"])
            
            kwargs = dict(
                date=date,
                desc=row["description"],
                amount=amount,
                type=transaction_type,
                extra={
                    "balance": decimal.Decimal(row["balance"])
                }
            )
            
            yield Transaction(
                extractor=self.EXTRACTOR_NAME,
                file=filename,
                lineno=i + 1,
                reversed_lineno=i - row_count,
                **kwargs
            )



import csv
import decimal
import hashlib
import os
import typing

import iso8601

from ..data_types import Fingerprint
from ..data_types import Transaction
from ..utils import parse_date
from .base import ExtractorBase

SIMPLE_VALUE_FIELDS = [
    "date",
    "name",
    "amount",
    "pending",
    "website",
    "datetime",
    "logo_url",
    "account_id",
    "category_id",
    "check_number",
    "account_owner",
    "merchant_name",
    "transaction_id",
    "authorized_date",
    "payment_channel",
    "transaction_code",
    "transaction_type",
    "iso_currency_code",
    "merchant_entity_id",
    "authorized_datetime",
    "pending_transaction_id",
    "unofficial_currency_code",
    "personal_finance_category_icon_url",
]
COUNTER_PARTIES_PREFIX = "counterparties__"
COUNTER_PARTIES_FIELDS = [
    "name",
    "type",
    "website",
    "logo_url",
    "entity_id",
    "phone_number",
    "confidence_level",
]
PERSONAL_FINANCE_CATEGORY_PREFIX = "personal_finance_category__"
PERSONAL_FINANCE_CATEGORY_FIELDS = ["primary", "detailed", "confidence_level"]
ALL_FIELDS = (
    SIMPLE_VALUE_FIELDS
    + list(map(lambda x: COUNTER_PARTIES_PREFIX + x, COUNTER_PARTIES_FIELDS))
    + list(
        map(
            lambda x: PERSONAL_FINANCE_CATEGORY_PREFIX + x,
            PERSONAL_FINANCE_CATEGORY_FIELDS,
        )
    )
)


class PlaidExtractor(ExtractorBase):
    EXTRACTOR_NAME = "plaid"
    DEFAULT_IMPORT_ID = "{{ transaction_id }}"

    def detect(self) -> bool:
        reader = csv.DictReader(self.input_file)
        try:
            return reader.fieldnames == ALL_FIELDS
        except Exception:
            return False

    def fingerprint(self) -> Fingerprint | None:
        reader = csv.DictReader(self.input_file)
        try:
            row = next(reader)
        except StopIteration:
            return
        hash = hashlib.sha256()
        for field in reader.fieldnames:
            hash.update(row[field].encode("utf8"))

        raw_authorized_date = row.pop("authorized_date")
        raw_date_value = row.pop("date")
        if raw_authorized_date.strip():
            date_value = raw_authorized_date
        else:
            date_value = raw_date_value
        return Fingerprint(
            starting_date=parse_date(date_value),
            first_row_hash=hash.hexdigest(),
        )

    def __call__(self) -> typing.Generator[Transaction, None, None]:
        filename = None
        if hasattr(self.input_file, "name"):
            filename = self.input_file.name
        row_count_reader = csv.DictReader(self.input_file)
        row_count = 0
        for _ in row_count_reader:
            row_count += 1
        self.input_file.seek(os.SEEK_SET, 0)
        reader = csv.DictReader(self.input_file)
        for i, row in enumerate(reader):
            pending = row.pop("pending").lower() == "true"
            if pending:
                date = parse_date(row.pop("date"))
                post_date = None
            else:
                raw_authorized_date = row.pop("authorized_date")
                date_value = parse_date(row.pop("date"))
                post_date = date_value
                if not raw_authorized_date.strip():
                    # in some strange situation, authorized_date could be empty, such as sandbox mode plaid credit card,
                    # not sure if this could also happen in production. but regardless, if it's the case, let's just
                    # use the date value as date and post date in the same time
                    date = date_value
                else:
                    date = parse_date(raw_authorized_date)

            dt = row.pop("datetime")
            if not dt:
                timestamp = None
            else:
                timestamp = iso8601.parse_date(dt)

            txn_id = row.pop("transaction_id")
            # For some banks, such as AMEX credit cards, when a pending transaction posted, the old one will
            # be deleted and a new one with the pending transaction id for the old one will be added.
            # To avoid txn id change after it gets posted, we should always use pending txn id first if available
            pending_transaction_id = row.pop("pending_transaction_id")
            if pending_transaction_id.strip():
                txn_id = pending_transaction_id
            kwargs = dict(
                transaction_id=txn_id,
                date=date,
                post_date=post_date,
                status="pending" if pending else "posted",
                pending=pending,
                desc=row.pop("name"),
                payee=row.pop("merchant_name"),
                source_account=row.pop("account_id"),
                amount=decimal.Decimal(row.pop("amount")),
                type=row.pop("payment_channel"),
                currency=row.pop("iso_currency_code"),
                category=row.pop("personal_finance_category__primary"),
                subcategory=row.pop("personal_finance_category__detailed"),
                timestamp=timestamp,
            )
            if row:
                kwargs["extra"] = row

            yield Transaction(
                extractor=self.EXTRACTOR_NAME,
                file=filename,
                lineno=i + 1,
                reversed_lineno=i - row_count,
                **kwargs,
            )



import pathlib

import pytest

from beanhub_extract.extractors import detect_extractor


@pytest.mark.parametrize(
    "input_file, extractor_name",
    [
        ("mercury.csv", "mercury"),
        ("chase_credit_card.csv", "chase_credit_card"),
        ("csv.csv", "csv"),
        ("wealthsimple.csv", "wealthsimple"),
        ("other.csv", None),
        (pytest.lazy_fixture("zip_file"), None),
    ],
)
def test_detect_extractor(
    fixtures_folder: pathlib.Path, input_file: str, extractor_name: str | None
):
    with open(fixtures_folder / input_file, "rt") as fo:
        extractor_cls = detect_extractor(fo)
        if extractor_name is None:
            assert extractor_cls is None
        else:
            assert extractor_cls.EXTRACTOR_NAME == extractor_name



import csv
import datetime
import decimal
import hashlib
import os
import typing
from dataclasses import fields

from ..data_types import Fingerprint
from ..data_types import Transaction
from ..utils import parse_date
from .base import ExtractorBase

EXCLUDED_FIELDS = frozenset(["extractor", "file", "lineno", "reversed_lineno", "extra"])
ALL_FIELDS = frozenset(field.name for field in fields(Transaction)) - EXCLUDED_FIELDS


class CSVExtractor(ExtractorBase):
    EXTRACTOR_NAME = "csv"
    DEFAULT_IMPORT_ID = "{{ file | as_posix_path }}:{{ lineno }}"

    def detect(self) -> bool:
        reader = csv.DictReader(self.input_file)
        try:
            return frozenset(reader.fieldnames).issuperset(ALL_FIELDS)
        except Exception:
            return False

    def fingerprint(self) -> Fingerprint | None:
        reader = csv.DictReader(self.input_file)
        try:
            row = next(reader)
        except StopIteration:
            return
        hash = hashlib.sha256()
        for field in reader.fieldnames:
            hash.update(row[field].encode("utf8"))

        date = row.pop("date")
        post_date = row.pop("post_date")
        if date.strip():
            date_value = date
        else:
            date_value = post_date
        if not date_value:
            date_value = datetime.date(1970, 1, 1)
        return Fingerprint(
            starting_date=parse_date(date_value),
            first_row_hash=hash.hexdigest(),
        )

    def __call__(self) -> typing.Generator[Transaction, None, None]:
        filename = None
        if hasattr(self.input_file, "name"):
            filename = self.input_file.name
        row_count_reader = csv.DictReader(self.input_file)
        row_count = 0
        for _ in row_count_reader:
            row_count += 1
        self.input_file.seek(os.SEEK_SET, 0)
        reader = csv.DictReader(self.input_file)
        for i, row in enumerate(reader):
            kwargs = {}
            for field in fields(Transaction):
                if field.name not in row:
                    continue
                value = row.pop(field.name)
                field_type, _ = typing.get_args(field.type)
                if value is not None and value.strip():
                    if field_type is datetime.date:
                        value = parse_date(value)
                    elif field_type is datetime.datetime:
                        value = datetime.datetime.fromisoformat(value)
                    elif field_type is decimal.Decimal:
                        value = decimal.Decimal(value)
                    elif field_type is bool:
                        value = value.lower() == "true"
                    kwargs[field.name] = value
            if row:
                kwargs["extra"] = row

            yield Transaction(
                extractor=self.EXTRACTOR_NAME,
                file=filename,
                lineno=i + 1,
                reversed_lineno=i - row_count,
                **kwargs,
            )



import pathlib
import zipfile

import pytest

TEST_PACKAGE_FOLDER = pathlib.Path(__file__).parent
FIXTURE_FOLDER = TEST_PACKAGE_FOLDER / "fixtures"


@pytest.fixture
def fixtures_folder() -> pathlib.Path:
    return FIXTURE_FOLDER


@pytest.fixture
def zip_file(tmp_path: pathlib.Path) -> pathlib.Path:
    text_file = tmp_path / "hello.txt"
    text_file.write_text("BeanHub is awesome")
    zip_file = tmp_path / "hello.zip"
    with zipfile.ZipFile(zip_file, "w") as zip:
        zip.write(text_file)
    return zip_file



import datetime
import decimal
import functools
import pathlib

import pytest

from beanhub_extract.data_types import Fingerprint
from beanhub_extract.data_types import Transaction
from beanhub_extract.extractors.plaid import PlaidExtractor
from beanhub_extract.utils import strip_txn_base_path


@pytest.mark.parametrize(
    "input_file, expected",
    [
        (
            "plaid.csv",
            [
                Transaction(
                    extractor="plaid",
                    file="plaid.csv",
                    lineno=1,
                    reversed_lineno=-6,
                    transaction_id="R55WBQZRx5h8J8dEKmPbFeda45RL6RCapLk7B",
                    date=datetime.date(2024, 1, 7),
                    post_date=datetime.date(2024, 1, 8),
                    desc="Uber 072515 SF**POOL**",
                    amount=decimal.Decimal("6.33"),
                    currency="USD",
                    category="TRANSPORTATION",
                    subcategory="TRANSPORTATION_TAXIS_AND_RIDE_SHARES",
                    status="posted",
                    pending=False,
                    type="online",
                    source_account="g554pPNvm5hG9GNgnzDpirvLMz5WaAcENmPM1",
                    payee="Uber",
                    extra={
                        "account_owner": "",
                        "authorized_datetime": "",
                        "category_id": "22016000",
                        "check_number": "",
                        "counterparties__confidence_level": "VERY_HIGH",
                        "counterparties__entity_id": "eyg8o776k0QmNgVpAmaQj4WgzW9Qzo6O51gdd",
                        "counterparties__logo_url": "https://plaid-merchant-logos.plaid.com/uber_1060.png",
                        "counterparties__name": "Uber",
                        "counterparties__phone_number": "",
                        "counterparties__type": "merchant",
                        "counterparties__website": "uber.com",
                        "logo_url": "https://plaid-merchant-logos.plaid.com/uber_1060.png",
                        "merchant_entity_id": "eyg8o776k0QmNgVpAmaQj4WgzW9Qzo6O51gdd",
                        "personal_finance_category__confidence_level": "VERY_HIGH",
                        "personal_finance_category_icon_url": "https://plaid-category-icons.plaid.com/PFC_TRANSPORTATION.png",
                        "transaction_code": "",
                        "transaction_type": "special",
                        "unofficial_currency_code": "",
                        "website": "uber.com",
                    },
                ),
                Transaction(
                    extractor="plaid",
                    file="plaid.csv",
                    lineno=2,
                    reversed_lineno=-5,
                    transaction_id="vPPEBM3p8PiAjAKZ1oRzCVEAwb7BK7iq6Pear",
                    date=datetime.date(2024, 1, 20),
                    post_date=datetime.date(2024, 1, 21),
                    desc="SparkFun",
                    amount=decimal.Decimal("89.4"),
                    currency="USD",
                    category="GENERAL_MERCHANDISE",
                    subcategory="GENERAL_MERCHANDISE_OTHER_GENERAL_MERCHANDISE",
                    status="posted",
                    pending=False,
                    type="in store",
                    source_account="g554pPNvm5hG9GNgnzDpirvLMz5WaAcENmPM1",
                    payee="FUN",
                    extra={
                        "account_owner": "",
                        "authorized_datetime": "",
                        "category_id": "13005000",
                        "check_number": "",
                        "counterparties__confidence_level": "LOW",
                        "counterparties__entity_id": "",
                        "counterparties__logo_url": "",
                        "counterparties__name": "FUN",
                        "counterparties__phone_number": "",
                        "counterparties__type": "merchant",
                        "counterparties__website": "",
                        "logo_url": "",
                        "merchant_entity_id": "",
                        "personal_finance_category__confidence_level": "LOW",
                        "personal_finance_category_icon_url": "https://plaid-category-icons.plaid.com/PFC_GENERAL_MERCHANDISE.png",
                        "transaction_code": "",
                        "transaction_type": "place",
                        "unofficial_currency_code": "",
                        "website": "",
                    },
                ),
                Transaction(
                    extractor="plaid",
                    file="plaid.csv",
                    lineno=3,
                    reversed_lineno=-4,
                    transaction_id="9AAqkd5G9AfRlRAp731jcGx1pAnM9nH4x3yzA",
                    date=datetime.date(2024, 1, 22),
                    post_date=datetime.date(2024, 1, 22),
                    desc="Starbucks",
                    amount=decimal.Decimal("4.33"),
                    currency="USD",
                    category="FOOD_AND_DRINK",
                    subcategory="FOOD_AND_DRINK_COFFEE",
                    status="posted",
                    pending=False,
                    type="in store",
                    source_account="g554pPNvm5hG9GNgnzDpirvLMz5WaAcENmPM1",
                    payee="Starbucks",
                    extra={
                        "account_owner": "",
                        "authorized_datetime": "",
                        "category_id": "13005043",
                        "check_number": "",
                        "counterparties__confidence_level": "VERY_HIGH",
                        "counterparties__entity_id": "NZAJQ5wYdo1W1p39X5q5gpb15OMe39pj4pJBb",
                        "counterparties__logo_url": "https://plaid-merchant-logos.plaid.com/starbucks_956.png",
                        "counterparties__name": "Starbucks",
                        "counterparties__phone_number": "",
                        "counterparties__type": "merchant",
                        "counterparties__website": "starbucks.com",
                        "logo_url": "https://plaid-merchant-logos.plaid.com/starbucks_956.png",
                        "merchant_entity_id": "NZAJQ5wYdo1W1p39X5q5gpb15OMe39pj4pJBb",
                        "personal_finance_category__confidence_level": "VERY_HIGH",
                        "personal_finance_category_icon_url": "https://plaid-category-icons.plaid.com/PFC_FOOD_AND_DRINK.png",
                        "transaction_code": "",
                        "transaction_type": "place",
                        "unofficial_currency_code": "",
                        "website": "starbucks.com",
                    },
                ),
                Transaction(
                    extractor="plaid",
                    file="plaid.csv",
                    lineno=4,
                    reversed_lineno=-3,
                    transaction_id="yeeGdMDA8eTWvWBMELzxcgGnVP7m67i46ZnkE",
                    date=datetime.date(2024, 1, 22),
                    post_date=datetime.date(2024, 1, 22),
                    desc="McDonald's",
                    amount=decimal.Decimal("12.0"),
                    currency="USD",
                    category="FOOD_AND_DRINK",
                    subcategory="FOOD_AND_DRINK_FAST_FOOD",
                    status="posted",
                    pending=False,
                    type="in store",
                    source_account="g554pPNvm5hG9GNgnzDpirvLMz5WaAcENmPM1",
                    payee="McDonald's",
                    extra={
                        "account_owner": "",
                        "authorized_datetime": "",
                        "category_id": "13005032",
                        "check_number": "",
                        "counterparties__confidence_level": "VERY_HIGH",
                        "counterparties__entity_id": "vzWXDWBjB06j5BJoD3Jo84OJZg7JJzmqOZA22",
                        "counterparties__logo_url": "https://plaid-merchant-logos.plaid.com/mcdonalds_619.png",
                        "counterparties__name": "McDonald's",
                        "counterparties__phone_number": "",
                        "counterparties__type": "merchant",
                        "counterparties__website": "mcdonalds.com",
                        "logo_url": "https://plaid-merchant-logos.plaid.com/mcdonalds_619.png",
                        "merchant_entity_id": "vzWXDWBjB06j5BJoD3Jo84OJZg7JJzmqOZA22",
                        "personal_finance_category__confidence_level": "VERY_HIGH",
                        "personal_finance_category_icon_url": "https://plaid-category-icons.plaid.com/PFC_FOOD_AND_DRINK.png",
                        "transaction_code": "",
                        "transaction_type": "place",
                        "unofficial_currency_code": "",
                        "website": "mcdonalds.com",
                    },
                ),
                Transaction(
                    extractor="plaid",
                    file="plaid.csv",
                    lineno=5,
                    reversed_lineno=-2,
                    transaction_id="KgmABN0aEXC3DZkRe5kQH3xVnwBkBmUv3jMz6",
                    date=datetime.date(2024, 1, 23),
                    post_date=datetime.date(2024, 1, 23),
                    desc="United Airlines",
                    amount=decimal.Decimal("-500.0"),
                    currency="USD",
                    category="TRAVEL",
                    subcategory="TRAVEL_FLIGHTS",
                    status="posted",
                    pending=False,
                    type="in store",
                    source_account="g554pPNvm5hG9GNgnzDpirvLMz5WaAcENmPM1",
                    payee="United Airlines",
                    extra={
                        "account_owner": "",
                        "authorized_datetime": "",
                        "category_id": "22001000",
                        "check_number": "",
                        "counterparties__confidence_level": "VERY_HIGH",
                        "counterparties__entity_id": "NKDjqyAdQQzpyeD8qpLnX0D6yvLe2KYKYYzQ4",
                        "counterparties__logo_url": "https://plaid-merchant-logos.plaid.com/united_airlines_1065.png",
                        "counterparties__name": "United Airlines",
                        "counterparties__phone_number": "",
                        "counterparties__type": "merchant",
                        "counterparties__website": "united.com",
                        "logo_url": "https://plaid-merchant-logos.plaid.com/united_airlines_1065.png",
                        "merchant_entity_id": "NKDjqyAdQQzpyeD8qpLnX0D6yvLe2KYKYYzQ4",
                        "personal_finance_category__confidence_level": "VERY_HIGH",
                        "personal_finance_category_icon_url": "https://plaid-category-icons.plaid.com/PFC_TRAVEL.png",
                        "transaction_code": "",
                        "transaction_type": "special",
                        "unofficial_currency_code": "",
                        "website": "united.com",
                    },
                ),
                Transaction(
                    extractor="plaid",
                    file="plaid.csv",
                    lineno=6,
                    reversed_lineno=-1,
                    transaction_id="ZrD8XbQvm7clZJe6EzMpcK5L3KebEMIe1xK5g",
                    date=datetime.date(2024, 1, 27),
                    post_date=datetime.date(2024, 1, 27),
                    timestamp=None,
                    timezone=None,
                    desc="Madison Bicycle Shop",
                    bank_desc=None,
                    amount=decimal.Decimal("500.0"),
                    currency="USD",
                    category="GENERAL_MERCHANDISE",
                    subcategory="GENERAL_MERCHANDISE_SPORTING_GOODS",
                    pending=False,
                    status="posted",
                    type="in store",
                    source_account="AMbx81AaBNixMg9kEy85caBeQnklzwI9JxMbP",
                    dest_account=None,
                    note=None,
                    reference=None,
                    payee="",
                    gl_code=None,
                    name_on_card=None,
                    last_four_digits=None,
                    extra={
                        "account_owner": "",
                        "authorized_datetime": "",
                        "category_id": "19046000",
                        "check_number": "",
                        "counterparties__confidence_level": "",
                        "counterparties__entity_id": "",
                        "counterparties__logo_url": "",
                        "counterparties__name": "",
                        "counterparties__phone_number": "",
                        "counterparties__type": "",
                        "counterparties__website": "",
                        "logo_url": "",
                        "merchant_entity_id": "",
                        "personal_finance_category__confidence_level": "LOW",
                        "personal_finance_category_icon_url": "https://plaid-category-icons.plaid.com/PFC_GENERAL_MERCHANDISE.png",
                        "transaction_code": "",
                        "transaction_type": "place",
                        "unofficial_currency_code": "",
                        "website": "",
                    },
                ),
            ],
        ),
    ],
)
def test_extractor(
    fixtures_folder: pathlib.Path, input_file: str, expected: list[Transaction]
):
    with open(fixtures_folder / input_file, "rt") as fo:
        extractor = PlaidExtractor(fo)
        assert (
            list(
                map(
                    functools.partial(strip_txn_base_path, fixtures_folder), extractor()
                )
            )
            == expected
        )


@pytest.mark.parametrize(
    "input_file, expected",
    [
        ("plaid.csv", True),
        ("mercury.csv", False),
        ("chase_credit_card.csv", False),
        ("empty.csv", False),
        ("other.csv", False),
        (pytest.lazy_fixture("zip_file"), False),
    ],
)
def test_detect(fixtures_folder: pathlib.Path, input_file: str, expected: bool):
    with open(fixtures_folder / input_file, "rt") as fo:
        extractor = PlaidExtractor(fo)
        assert extractor.detect() == expected


def test_fingerprint(fixtures_folder: pathlib.Path):
    with open(fixtures_folder / "plaid.csv", "rt") as fo:
        extractor = PlaidExtractor(fo)
        assert extractor.fingerprint() == Fingerprint(
            starting_date=datetime.date(2024, 1, 7),
            first_row_hash="264b70b31bf27298bef4357774ed2ca09f4a4e64e7e73cfb985b3afff76b5b06",
        )



import dataclasses
import datetime
import pathlib

from .data_types import Transaction


def strip_base_path(
    base: pathlib.Path | pathlib.PurePath,
    filepath: str | pathlib.Path | pathlib.PurePath,
    pure_posix: bool = False,
) -> str:
    """Strip file base path (parent folder) from given file path"""
    if not isinstance(filepath, pathlib.Path):
        if pure_posix:
            filepath = pathlib.PurePosixPath(filepath)
        else:
            filepath = pathlib.Path(filepath)
    return str(filepath.relative_to(base))


def strip_txn_base_path(
    base: pathlib.Path | pathlib.PurePath,
    transaction: Transaction,
    pure_posix: bool = False,
) -> Transaction:
    """Strip file base path (parent folder) from given transaction"""
    if transaction.file is None:
        return transaction
    return Transaction(
        **(
            dataclasses.asdict(transaction)
            | dict(file=strip_base_path(base, transaction.file, pure_posix))
        )
    )


def parse_date(date_str: str) -> datetime.date:
    parts = date_str.split("-")
    return datetime.date(*(map(int, parts)))



import os
import typing

from .base import ExtractorBase
from .chase import ChaseCreditCardExtractor
from .csv import CSVExtractor
from .mercury import MercuryExtractor
from .plaid import PlaidExtractor
from .wealthsimple import WealthsimpleExtractor

ALL_EXTRACTORS: dict[str, typing.Type[ExtractorBase]] = {
    MercuryExtractor.EXTRACTOR_NAME: MercuryExtractor,
    ChaseCreditCardExtractor.EXTRACTOR_NAME: ChaseCreditCardExtractor,
    PlaidExtractor.EXTRACTOR_NAME: PlaidExtractor,
    WealthsimpleExtractor.EXTRACTOR_NAME: WealthsimpleExtractor,
    CSVExtractor.EXTRACTOR_NAME: CSVExtractor,
}


def detect_extractor(input_file: typing.TextIO) -> typing.Type[ExtractorBase]:
    for extractor_cls in ALL_EXTRACTORS.values():
        input_file.seek(os.SEEK_SET)
        if extractor_cls(input_file).detect():
            return extractor_cls
